searchNodes=[{"doc":"Top-level module holding the Packmatic library, which provides Zip-oriented stream aggregation services from various sources.","ref":"Packmatic.html","title":"Packmatic","type":"module"},{"doc":"Builds a Stream which can be consumed to construct a Zip file from various sources, as specified in the Manifest. When building the Stream, options can be passed to configure how the Encoder should behave when Source acquisition fails. Examples The Stream can be created by passing a Packmatic.Manifest.t/0 struct, a list of Manifest Entries ( Packmatic.Manifest.Entry.t/0 ), or a list of Keyword Lists that are understood and can be transformed to Manifest Entries ( Packmatic.Manifest.Entry.proplist/0 ). iex(1)&gt; stream = Packmatic . build_stream ( Packmatic.Manifest . create ( ) ) iex(2)&gt; is_function ( stream ) true iex(1)&gt; stream = Packmatic . build_stream ( [ ] ) iex(2)&gt; is_function ( stream ) true iex(1)&gt; stream = Packmatic . build_stream ( [ [ source : { :file , &quot;foo.bar&quot; } ] ] ) iex(2)&gt; is_function ( stream ) true","ref":"Packmatic.html#build_stream/2","title":"Packmatic.build_stream/2","type":"function"},{"doc":"Holds logic which can be used to put together a Zip file in an interative fashion, suitable for wrapping within a Stream . The format of Zip files emitted by Packmatic.Encoder is documented under the modules implementing the Packmatic.Field protocol. The Encoder is wrapped in Stream.resource/3 for consumption as an Elixir Stream, under Packmatic.build_stream/1 . Further, the Stream can be used with Plug.Conn to serve a chunked connection easily, as provided in Packmatic.Conn.send_chunked/3 . The Encoder has three statuses: Encoding, where each Entry within the Manifest is transformed to a Source, which is subsequently consumed. If the on_error option is set to :skip when building the stream, then sources which have raised error are skipped, although at this time portions of the source may have already been sent. Otherwise, and as the default behaviour, an uncaught exception will be raised and the consumer of the Stream will crash. During Encoding, content is dynamically deflated. Journaling, where each successfully encoded Entry is journaled again at the end of the archive, with the Central Directory structure. Both Zip and Zip64 formats are used for maximum flexibility. In case the on_error option is set to :skip , any source which has raised an error during its consumption will not be journaled. Due to the nature of streaming archives, this may still leave portions of unusable data within the archive. Done, which is the terminal status.","ref":"Packmatic.Encoder.html","title":"Packmatic.Encoder","type":"module"},{"doc":"Completes the Stream.","ref":"Packmatic.Encoder.html#stream_after/2","title":"Packmatic.Encoder.stream_after/2","type":"function"},{"doc":"Iterates the Stream. When the Stream is in :encoding status, this function may continue encoding of the current item, or advance to the next item, or advance to the :journaling status when there are no further items to encode. When the Stream is in :journaling status, this function may continue journaling the next item, or advance to the :done status. When the Stream is in :done status, it can not be iterated further.","ref":"Packmatic.Encoder.html#stream_next/2","title":"Packmatic.Encoder.stream_next/2","type":"function"},{"doc":"Starts the Stream. If the Manifest provided is invalid, the call will not succeed and the invalid Manifest will be returned.","ref":"Packmatic.Encoder.html#stream_start/2","title":"Packmatic.Encoder.stream_start/2","type":"function"},{"doc":"Represents the intenral state used when encoding entries.","ref":"Packmatic.Encoder.html#t:encoding_state/0","title":"Packmatic.Encoder.encoding_state/0","type":"opaque"},{"doc":"Represents the internal state used when journaling entries.","ref":"Packmatic.Encoder.html#t:journaling_state/0","title":"Packmatic.Encoder.journaling_state/0","type":"opaque"},{"doc":"Represents possible options to use with the Encoder. on_error can be set to either :skip or :halt , which controls how the Encoder behaves when there is an error with one of the Sources. on_event can be set to a function which will be called when events are raised by the Encoder during its lifecycle. See Packmatic.Event for further information.","ref":"Packmatic.Encoder.html#t:option/0","title":"Packmatic.Encoder.option/0","type":"type"},{"doc":"Represents an unique identifier of the Stream in operation. This allows you to distinguish between multiple series of Events raised against the same Manifest in multiple Streams concurrently.","ref":"Packmatic.Encoder.html#t:stream_id/0","title":"Packmatic.Encoder.stream_id/0","type":"opaque"},{"doc":"Represents the customer’s request for a particular compressed file. The Manifest is constructed with a list of Packmatic.Manifest.Entry structs, which each represents one file to be placed into the Package. Entries are validated when they are added to the Manifest. Creating Manifests Manifests can be created iteratively by calling prepend/2 against an existing Manifest, or by calling create/1 with a list of Entries created elsewhere. Calling create/0 provides you with an empty Manifest which is not valid: iex(1)&gt; Packmatic.Manifest . create ( ) % Packmatic.Manifest { entries : [ ] , errors : [ manifest : :empty ] , valid? : false } However, prepending valid Entries makes it valid: iex(1)&gt; manifest = Packmatic.Manifest . create ( ) iex(2)&gt; manifest = Packmatic.Manifest . prepend ( manifest , source : { :random , 1 } , path : &quot;foo&quot; ) iex(3)&gt; manifest . valid? true Creating a Manifest from a list of Entries also results in a valid Manifest: iex(1)&gt; manifest = Packmatic.Manifest . create ( [ [ source : { :random , 1 } , path : &quot;foo&quot; ] ] ) iex(2)&gt; manifest . valid? true Validity and Error Reporting An empty Manifest is not valid because the result is not useful. It contains a Manifest-level error {:manifest, :empty} . iex(1)&gt; manifest = Packmatic.Manifest . create ( ) iex(2)&gt; manifest . errors [ manifest : :empty ] If an Entry is not valid when appended to a Manifest, it will make the Manifest invalid and a corresponding error entry will also be added to the :errors key. Since entries are prepended to the list, Entry-level errors are emitted with a negative index (counted from the tail of the list). So the last item (first to be appended) has the index of -1 , the penultimate item has the index of -2 , and so on. iex(1)&gt; manifest = Packmatic.Manifest . create ( ) iex(2)&gt; manifest = Packmatic.Manifest . prepend ( manifest , source : { :file , nil } ) iex(3)&gt; manifest . errors [ { { :entry , - 1 } , [ source : :invalid , path : :missing ] } ] Entry-level errors are emitted per key, and defined under Packmatic.Manifest.Entry.error/0 .","ref":"Packmatic.Manifest.html","title":"Packmatic.Manifest","type":"module"},{"doc":"Creates a Manifest based on Entries given. If there are no Entries, the Manifest will be invalid by default. Otherwise, each Entry will be validated and the Manifest will remain valid if all Entries provided were valid.","ref":"Packmatic.Manifest.html#create/1","title":"Packmatic.Manifest.create/1","type":"function"},{"doc":"Prepends the given Entry to the Manifest. If the Entry is invalid, the Manifest will also become invalid, and the error will be prepended to the list.","ref":"Packmatic.Manifest.html#prepend/2","title":"Packmatic.Manifest.prepend/2","type":"function"},{"doc":"Represents an Error which can be related to an Entry or the Manifest itself.","ref":"Packmatic.Manifest.html#t:error/0","title":"Packmatic.Manifest.error/0","type":"type"},{"doc":"Represents an Error regarding an Entry. Index is negative, counted from tail.","ref":"Packmatic.Manifest.html#t:error_entry/0","title":"Packmatic.Manifest.error_entry/0","type":"type"},{"doc":"Represents an Error with the Manifest.","ref":"Packmatic.Manifest.html#t:error_manifest/0","title":"Packmatic.Manifest.error_manifest/0","type":"type"},{"doc":"Represents an error condition where the Manifest has no entries.","ref":"Packmatic.Manifest.html#t:error_manifest_reason/0","title":"Packmatic.Manifest.error_manifest_reason/0","type":"type"},{"doc":"Represents an invalid Manifest, where there must be errors and might be Entries.","ref":"Packmatic.Manifest.html#t:invalid/0","title":"Packmatic.Manifest.invalid/0","type":"type"},{"doc":"Represents the Manifest.","ref":"Packmatic.Manifest.html#t:t/0","title":"Packmatic.Manifest.t/0","type":"type"},{"doc":"Represents a valid Manifest, where there must be Entries and no errors.","ref":"Packmatic.Manifest.html#t:valid/0","title":"Packmatic.Manifest.valid/0","type":"type"},{"doc":"Defines how data can be acquired in a piecemeal fashion, perhaps by reading only a few pages from the disk at a time or only a few MBs of data from an open socket. The Source behaviour defines three callbacks that must be implemented by conforming modules: validate/1 , which is called to check the initialisation argument. init/1 , which is called to instantiate the source and return its state. read/1 , which is called to read data from the source, given the state. Representing Sources Sources are represented in Manifest Entries as tuples such as {:file, path} or {:url, url} . This form of representation is called a Source Entry. The Source Entry is a stable locator of the underlying data which has no runtime implications. The Encoder hydrates the Source Entry into whatever the Source module implements internally, when it is time to pull data from that source. The first element in the tuple is the Source Name, and the second element is called the Initialisation Argument ( init_arg ). Source Name The Source names can be special atoms (short names) or full module names: :file resolves to Packmatic.Source.File . :url resolves to Packmatic.Source.URL . :dynamic resolves to Packmatic.Source.Dynamic . :random resolves to Packmatic.Source.Random . When using Custom Sources, the module name can be passed (as an atom) as well. Initialisation Argument The Initialisation Argument is usually a basic Elixir type, but in the case of Dynamic Sources, it is a function which resolves to a Source Entry understood by either the File or URL source. Examples The Source Entry {:file, path} is resolved during encoding: iex(1)&gt; { :ok , file_path } = Briefly . create ( ) iex(2)&gt; { :ok , { module , state } } = Packmatic.Source . build ( { :file , file_path } ) iex(3)&gt; module Packmatic.Source.File iex(3)&gt; state . __struct__ Packmatic.Source.File Notes When implementing a custom Source which uses an external data provider (for example reading from a file), remember to perform any cleanup required within the read/1 callback if the Source is not expected to return any further data, for example if the file has been read completely or if there has been an error.","ref":"Packmatic.Source.html","title":"Packmatic.Source","type":"behaviour"},{"doc":"Initialises the Source with the Initialisation Argument as specified in the Entry. This prepares the Source for acquisition. Called by Packmatic.Encoder .","ref":"Packmatic.Source.html#build/1","title":"Packmatic.Source.build/1","type":"function"},{"doc":"Converts the Entry to a Source State.","ref":"Packmatic.Source.html#c:init/1","title":"Packmatic.Source.init/1","type":"callback"},{"doc":"Consumes bytes off an initialised Source. Called by Packmatic.Encoder .","ref":"Packmatic.Source.html#read/1","title":"Packmatic.Source.read/1","type":"function"},{"doc":"Iterates the Source State. Usually this will return an IO List, or {:error, reason} . For stateful Sources, an updated Source State can be returned with the data. Note that Sources that have returned :eof or {:error, reason} will not be touched again.","ref":"Packmatic.Source.html#c:read/1","title":"Packmatic.Source.read/1","type":"callback"},{"doc":"Validates the given Entry. Called by Packmatic.Manifest.Entry .","ref":"Packmatic.Source.html#validate/1","title":"Packmatic.Source.validate/1","type":"function"},{"doc":"Validates the given Initialisation Argument.","ref":"Packmatic.Source.html#c:validate/1","title":"Packmatic.Source.validate/1","type":"callback"},{"doc":"Represents an internal tuple that can be used to initialise a Source with build/1 . This allows the Entries to be dynamically resolved. Dynamic sources use this to prepare their work lazily, and other Sources may use this mechanism to delay opening of sockets or handles.","ref":"Packmatic.Source.html#t:entry/0","title":"Packmatic.Source.entry/0","type":"type"},{"doc":"Represents the Initialisation Argument which is a stable locator for the underlying data, that the Source will initialise based upon.","ref":"Packmatic.Source.html#t:init_arg/0","title":"Packmatic.Source.init_arg/0","type":"type"},{"doc":"Represents the Name of the Source, which can be a shorthand (atom) or a module.","ref":"Packmatic.Source.html#t:name/0","title":"Packmatic.Source.name/0","type":"type"},{"doc":"Represents the internal State for a resolved Source that is being read from. In case of a File source, the state may be a struct which holds the File Handle; in case of a URL source, it may be the underlying network socket, etc.","ref":"Packmatic.Source.html#t:state/0","title":"Packmatic.Source.state/0","type":"type"},{"doc":"Represents a particular file that will go into package, which is sourced by reading from a file, downloading from an URI, etc. The source in the Manifest Entry is a Source Entry ( Packmatic.Source.entry/0 ), which will be dynamically resolved at runtime using Packmatic.Source.build/1 by the Encoder, when it is time to start reading from it.","ref":"Packmatic.Manifest.Entry.html","title":"Packmatic.Manifest.Entry","type":"module"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:error/0","title":"Packmatic.Manifest.Entry.error/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:error_path/0","title":"Packmatic.Manifest.Entry.error_path/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:error_source/0","title":"Packmatic.Manifest.Entry.error_source/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:error_timestamp/0","title":"Packmatic.Manifest.Entry.error_timestamp/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:path/0","title":"Packmatic.Manifest.Entry.path/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:proplist/0","title":"Packmatic.Manifest.Entry.proplist/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:source/0","title":"Packmatic.Manifest.Entry.source/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:t/0","title":"Packmatic.Manifest.Entry.t/0","type":"type"},{"doc":"","ref":"Packmatic.Manifest.Entry.html#t:timestamp/0","title":"Packmatic.Manifest.Entry.timestamp/0","type":"type"},{"doc":"Represents content which may be generated on-demand, for example by another subsystem or via downloading from a signed URL. For example, a function which dynamically generates a URL (perhaps a signed S3 URL in your own use case) would look like this: iex(1)&gt; url = &quot;https://example.com&quot; iex(2)&gt; init_arg = fn -&gt; { :ok , { :url , url } } end iex(3)&gt; { :ok , { module , source } } = Packmatic.Source.Dynamic . init ( init_arg ) iex(4)&gt; module Packmatic.Source.URL iex(5)&gt; source . __struct__ Packmatic.Source.URL And when used within a Manifest, it would look like this: iex(1)&gt; url = &quot;https://example.com&quot; iex(2)&gt; init_arg = fn -&gt; { :ok , { :url , url } } end iex(3)&gt; entry = [ source : { :dynamic , init_arg } , path : &quot;foo.pdf&quot; ] iex(4)&gt; manifest = Packmatic.Manifest . create ( [ entry ] ) iex(5)&gt; manifest . valid? true Even if the function (referenced by the Initialisation Argument) resolves cleanly, the result may still be rejected by the underlying Source, for example if the file does not exist. This kind of error “bubbles up” and is dealt with by the Encoder at runtime. iex(1)&gt; Packmatic.Source.Dynamic . init ( fn -&gt; { :ok , { :file , &quot;example.pdf&quot; } } end ) { :error , :enoent } However, since resolution happens only when the built Stream starts to be consumed, such a Source Entry would be valid when placed in a Manifest ahead of time: iex(1)&gt; path = &quot;example.pdf&quot; iex(2)&gt; init_arg = fn -&gt; { :ok , { :file , path } } end iex(3)&gt; entry = [ source : { :dynamic , init_arg } , path : &quot;foo.pdf&quot; ] iex(4)&gt; manifest = Packmatic.Manifest . create ( [ entry ] ) iex(5)&gt; manifest . valid? true","ref":"Packmatic.Source.Dynamic.html","title":"Packmatic.Source.Dynamic","type":"module"},{"doc":"","ref":"Packmatic.Source.Dynamic.html#t:init_arg/0","title":"Packmatic.Source.Dynamic.init_arg/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Dynamic.html#t:init_result/0","title":"Packmatic.Source.Dynamic.init_result/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Dynamic.html#t:resolve_fun/0","title":"Packmatic.Source.Dynamic.resolve_fun/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Dynamic.html#t:resolve_result/0","title":"Packmatic.Source.Dynamic.resolve_result/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Dynamic.html#t:resolve_result_error/0","title":"Packmatic.Source.Dynamic.resolve_result_error/0","type":"type"},{"doc":"Represents content on disk, for example from a static file. Also useful for content generated ahead of time.","ref":"Packmatic.Source.File.html","title":"Packmatic.Source.File","type":"module"},{"doc":"","ref":"Packmatic.Source.File.html#t:init_arg/0","title":"Packmatic.Source.File.init_arg/0","type":"type"},{"doc":"","ref":"Packmatic.Source.File.html#t:init_result/0","title":"Packmatic.Source.File.init_result/0","type":"type"},{"doc":"","ref":"Packmatic.Source.File.html#t:t/0","title":"Packmatic.Source.File.t/0","type":"type"},{"doc":"Represents randomly generated content, which is used mostly for testing, when you want to have a particular entry generated to a specific length.","ref":"Packmatic.Source.Random.html","title":"Packmatic.Source.Random","type":"module"},{"doc":"","ref":"Packmatic.Source.Random.html#t:init_arg/0","title":"Packmatic.Source.Random.init_arg/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Random.html#t:init_result/0","title":"Packmatic.Source.Random.init_result/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Random.html#t:t/0","title":"Packmatic.Source.Random.t/0","type":"type"},{"doc":"Represents content generated by enumerating a Stream, which returns IO Lists. Any data type that implements Enumerable can be used as the Initialisation Argument for this Source. Usually this would be a Stream that you have created elsewhere. In case the data type needs to be dynamically generated, you can instead use a Dynamic source, i.e. Packmatic.Source.Dynamic , and build the actual enum there.","ref":"Packmatic.Source.Stream.html","title":"Packmatic.Source.Stream","type":"module"},{"doc":"","ref":"Packmatic.Source.Stream.html#t:init_arg/0","title":"Packmatic.Source.Stream.init_arg/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Stream.html#t:init_result/0","title":"Packmatic.Source.Stream.init_result/0","type":"type"},{"doc":"","ref":"Packmatic.Source.Stream.html#t:t/0","title":"Packmatic.Source.Stream.t/0","type":"type"},{"doc":"Represents content which can be acquired by downloading from a remote server via HTTP(S) in chunks. Each chunk is then pulled away by the Encoder, which is iterated by the Stream. The underlying implementation is achieved via HTTPoison and Hackney.","ref":"Packmatic.Source.URL.html","title":"Packmatic.Source.URL","type":"module"},{"doc":"","ref":"Packmatic.Source.URL.html#build_options/1","title":"Packmatic.Source.URL.build_options/1","type":"function"},{"doc":"","ref":"Packmatic.Source.URL.html#init/3","title":"Packmatic.Source.URL.init/3","type":"function"},{"doc":"","ref":"Packmatic.Source.URL.html#t:headers/0","title":"Packmatic.Source.URL.headers/0","type":"type"},{"doc":"","ref":"Packmatic.Source.URL.html#t:init_arg/0","title":"Packmatic.Source.URL.init_arg/0","type":"type"},{"doc":"","ref":"Packmatic.Source.URL.html#t:init_result/0","title":"Packmatic.Source.URL.init_result/0","type":"type"},{"doc":"","ref":"Packmatic.Source.URL.html#t:options/0","title":"Packmatic.Source.URL.options/0","type":"type"},{"doc":"","ref":"Packmatic.Source.URL.html#t:t/0","title":"Packmatic.Source.URL.t/0","type":"type"},{"doc":"","ref":"Packmatic.Source.URL.html#t:target/0","title":"Packmatic.Source.URL.target/0","type":"type"},{"doc":"Represents Events that can be raised during the lifecycle of a Packmatic Stream being consumed. To listen for Events, you must pass a function reference to the on_event: option when calling Packmatic.build_stream/2 . This function will be called at appropriate junctures in the lifecycle of a Stream being consumed. The Events will be called in the following order: Packmatic.Event.StreamStarted : Sent when the Stream starts encoding. Packmatic.Event.EntryStarted : Sent when the Stream starts encoding data for a new entry. Packmatic.Event.EntryUpdated : Sent when the entry has initialised and some data has been read. This event will be sent on each iteration of Packmatic.Source.read/1 . Packmatic.Event.EntryFailed : Sent when the entry has failed to initialise (its Source returned an error during Packmatic.Source.init/1 ), in which case there would have been no EntryUpdated events, or when the entry has failed during the course of reading (its Source returned an error during Packmatic.Source.read/1 ). Packmatic.Event.EntryCompleted : Sent when the entry has been fully encoded (its Source has returned EOF). Packmatic.Event.StreamEnded : Sent when the Stream has completed journaling. Please note that more event types may be added in the future.","ref":"Packmatic.Event.html","title":"Packmatic.Event","type":"module"},{"doc":"Represents an Event that will be passed to the handler.","ref":"Packmatic.Event.html#t:event/0","title":"Packmatic.Event.event/0","type":"type"},{"doc":"Represents the callback function passed to the Encoder. The callback function takes 1 argument, which is the actual Event that is raised by Packmatic. The Event, event/0 , is one of the pre-defined structs under the Packmatic.Event namespace. Please keep in mind that more events may be added in the future, so you should always include a fallback clause in your handler function. Handlers are called from the same process that the Stream is being iterated from, which allows you to control what happens to it. Should you not wish to interrupt the Encoder, return :ok . Otherwise, if you must, you may raise an exception, which will crash the Stream.","ref":"Packmatic.Event.html#t:handler_fun/0","title":"Packmatic.Event.handler_fun/0","type":"type"},{"doc":"Represents an Event that is raised when the Encoder has completed reading from the Entry, i.e. when the Source has returned EOF.","ref":"Packmatic.Event.EntryCompleted.html","title":"Packmatic.Event.EntryCompleted","type":"module"},{"doc":"","ref":"Packmatic.Event.EntryCompleted.html#t:t/0","title":"Packmatic.Event.EntryCompleted.t/0","type":"type"},{"doc":"Represents an Event that is raised when the Encoder has failed to read from the Entry, i.e. when the Source has returned an error. Depending on the options, the Stream may continue to encode or it may halt.","ref":"Packmatic.Event.EntryFailed.html","title":"Packmatic.Event.EntryFailed","type":"module"},{"doc":"","ref":"Packmatic.Event.EntryFailed.html#t:t/0","title":"Packmatic.Event.EntryFailed.t/0","type":"type"},{"doc":"Represents an Event that is raised when the Encoder starts reading from a new Entry. The Entry is the same one as passed in the Manifest.","ref":"Packmatic.Event.EntryStarted.html","title":"Packmatic.Event.EntryStarted","type":"module"},{"doc":"","ref":"Packmatic.Event.EntryStarted.html#t:t/0","title":"Packmatic.Event.EntryStarted.t/0","type":"type"},{"doc":"Represents an Event that is raised when the Encoder has made progress reading from the Entry. Usually, the Source will read iteratively so this messge can be raised quite frequently.","ref":"Packmatic.Event.EntryUpdated.html","title":"Packmatic.Event.EntryUpdated","type":"module"},{"doc":"","ref":"Packmatic.Event.EntryUpdated.html#t:t/0","title":"Packmatic.Event.EntryUpdated.t/0","type":"type"},{"doc":"Represents an Event that is raised when the Encoder completes work. In case of normal completion, the reason will be set to :done , otherwise and in case of Source errors, the reason will be carried across.","ref":"Packmatic.Event.StreamEnded.html","title":"Packmatic.Event.StreamEnded","type":"module"},{"doc":"","ref":"Packmatic.Event.StreamEnded.html#t:t/0","title":"Packmatic.Event.StreamEnded.t/0","type":"type"},{"doc":"Represents an Event that is raised when a new copy of Encoder, and therefore a new Stream, is started against the Manifest.","ref":"Packmatic.Event.StreamStarted.html","title":"Packmatic.Event.StreamStarted","type":"module"},{"doc":"","ref":"Packmatic.Event.StreamStarted.html#t:t/0","title":"Packmatic.Event.StreamStarted.t/0","type":"type"},{"doc":"Represents data fields used internally by Packmatic.Encoder to represent information which make up the Zip format.","ref":"Packmatic.Field.html","title":"Packmatic.Field","type":"protocol"},{"doc":"Encodes the given structure into an IO List, or crashes if the structure is invalid.","ref":"Packmatic.Field.html#encode/1","title":"Packmatic.Field.encode/1","type":"function"},{"doc":"All the types that implement this protocol.","ref":"Packmatic.Field.html#t:t/0","title":"Packmatic.Field.t/0","type":"type"},{"doc":"Represents the End of Central Directory record. Within this implementation, the Zip64 standard always adopted, so the Zip64 End of Central Directory Record, and the Zip64 End of Central Directory Record Locator, are also emitted. Structure Zip64 End of Central Directory Record Size Content 4 bytes Signature 8 bytes Size of Record (excluding leading 12 bytes) 2 bytes Version made by 2 bytes Version needed to extract 4 bytes Number of this disk 4 bytes Number of the disk with the start of the Central Directory 8 bytes Total number of entries in the Central Directory on this disk 8 bytes Total number of entries in the Central Directory 8 bytes Size of the Central Directory 8 bytes Offset of start of Central Directory with respect to the starting disk number Variable Zip64 extensible data sector (variable, but empty in this implementation) Zip64 End of Central Directory Locator Size Content 4 bytes Signature 4 bytes Number of the disk with the start of the Zip64 End of Central Directory 8 bytes Relative offset of the Zip64 End of Central Directory record 4 bytes Total number of disks End of Central Directory Record Size Content 4 bytes Signature 2 bytes Number of this disk 2 bytes Number of the disk with the start of the Central Directory 2 bytes Total number of entries in the Central Directory on this disk 2 bytes Total number of entries in the Central Directory 4 bytes Size of the Central Directory 4 bytes Offset of start of Central Directory with respect to the starting disk number 2 bytes File comment length Variable File comment (64KB max.) Notes File comments are not emitted by Packmatic.","ref":"Packmatic.Field.Central.DirectoryEnd.html","title":"Packmatic.Field.Central.DirectoryEnd","type":"module"},{"doc":"","ref":"Packmatic.Field.Central.DirectoryEnd.html#t:t/0","title":"Packmatic.Field.Central.DirectoryEnd.t/0","type":"type"},{"doc":"Represents the Central Directory File Header, which is part of the Central Directory that is emitted after all successfully encoded files have been incorporated into the Zip stream. Structure Central Directory File Header Size Content 4 bytes Signature 1 byte Version made by - Zip Specification Version 1 byte Version made by - Environment 2 bytes Version needed to extract 2 bytes General Purpose Flag 2 bytes Compression Method (0 = No Compression; 8 = Deflated) 2 bytes Modification Time (DOS Format) 2 bytes Modification Date (DOS Format) 4 bytes Checksum (CRC-32) 4 bytes Compressed Size (Placeholder; value set in Zip64 Extended Information Extra Field) 4 bytes Original Size (Placeholder; value set in Zip64 Extended Information Extra Field) 2 bytes File Path Length (Bytes) 2 bytes Extra Fields Length (Bytes) 2 bytes File Comment Length (Bytes) 2 bytes Starting Disk Number for File 2 bytes Internal Attrbutes 4 bytes External Attrbutes 4 bytes Offset of Local File Header Variable File Path Variable Extra Fields Variable File Comment Notes The General Purpose Flag has the following bits set. Bit 3: Indicating a Streaming Archive; Data Descriptor is used, and the Local File Header has no Size or CRC information. Bit 11: Language encoding flag, indicating that the Filename and Comment are both already in UTF-8. As per APPNOTE, the presence of this flag obviates the need to emit a separate Info-ZIP Unicode Path Extra Field. The Compressed Size and Original Size fields are both set to 0xFF 0xFF 0xFF 0xFF , in order to force the real sizes, set in the Zip64 Extended Information Extra Field (provided by Packmatic.Field.Shared.ExtendedInformation ) to be used. The following Extra Fields are emitted: Extended Timestamp, see Packmatic.Field.Shared.ExtendedTimestamp Zip64 Extended Information, see Packmatic.Field.Shared.ExtendedInformation File comments are not emitted by Packmatic.","ref":"Packmatic.Field.Central.FileHeader.html","title":"Packmatic.Field.Central.FileHeader","type":"module"},{"doc":"","ref":"Packmatic.Field.Central.FileHeader.html#t:t/0","title":"Packmatic.Field.Central.FileHeader.t/0","type":"type"},{"doc":"Represents the Data Descriptor, which is used to facilitate streaming. This is requried since Packmatic assembles the files on the fly, so it does not know the size until the entire source has been read. Structure Data Descriptor Size Content 4 bytes Signature 4 bytes Checksum (CRC-32) 4 bytes Compressed Size (Bytes) 4 bytes Original Size (Bytes) Notes Although the APPNOTE indicates that Zip64 format should be used, 8-byte sizes crash the Unarchiver process on macOS High Sierra, but a truncated one works totally fine.","ref":"Packmatic.Field.Local.DataDescriptor.html","title":"Packmatic.Field.Local.DataDescriptor","type":"module"},{"doc":"","ref":"Packmatic.Field.Local.DataDescriptor.html#t:t/0","title":"Packmatic.Field.Local.DataDescriptor.t/0","type":"type"},{"doc":"Represents the Local File Header, which is emitted before the content of each file is incorporated into the Zip stream. Structure Local File Header Size Content 4 bytes Signature 2 bytes Version needed to extract 2 bytes General Purpose Flag 2 bytes Compression Method (0 = No Compression; 8 = Deflated) 2 bytes Modification Time (DOS Format) 2 bytes Modification Date (DOS Format) 4 bytes Checksum (CRC-32; 0 since Data Descriptor is used) 4 bytes Compressed Size (Bytes; 0 since Data Descriptor is used) 4 bytes Original Size (Bytes; 0 since Data Descriptor is used) 2 bytes File Path Length (Bytes) 2 bytes Extra Fields Length (Bytes) Variable File Path Variable Extra Fields Notes The General Purpose Flag has the following bits set. Bit 3: Indicating a Streaming Archive; Data Descriptor is used, and the Local File Header has no Size or CRC information. Bit 11: Language encoding flag, indicating that the Filename and Comment are both already in UTF-8. As per APPNOTE, the presence of this flag obviates the need to emit a separate Info-ZIP Unicode Path Extra Field. The Checksum, Compressed Size and Original Size fields are set to 0, since when the Local File Header is written, no further data has been read and so this information is not available. When the file has been read fully, a Data Descriptor will be written, which contains relevant information. The following Extra Field is emitted: Extended Timestamp, see Packmatic.Field.Shared.ExtendedTimestamp","ref":"Packmatic.Field.Local.FileHeader.html","title":"Packmatic.Field.Local.FileHeader","type":"module"},{"doc":"","ref":"Packmatic.Field.Local.FileHeader.html#t:t/0","title":"Packmatic.Field.Local.FileHeader.t/0","type":"type"},{"doc":"Represents the Zip64 Extended Information Extra Field, which can be emitted in both Local and Central File Headers, but in practice only used in the Central File Header within Packmatic, due to its streaming nature. Furthermore, disregarding the APPNOTE’s indication that the field should only be used if the sizes are set to 0xFF 0xFF / 0xFF 0xFF 0xFF 0xFF , since Packmatic always skips the file sizes in the Local File Header (due to the archive being generated in a streaming fashionn), and always emits the sizes in the Central Directory File Header as 0xFF 0xFF 0xFF 0xFF for simplicity, the Zip64 Extended Information Extra Field is never emitted by Packmatic in the Local File Header, and always emitted by Packmatic in the Central Directory File Header, with both Uncompressed and Compressed Sizes. Structure Shared Zip64 Extended Information Size Content 2 bytes Signature 2 bytes Size of Rest of Field (Bytes) 8 bytes Original Size (Bytes) 8 bytes Compressed Size (Bytes)","ref":"Packmatic.Field.Shared.ExtendedInformation.html","title":"Packmatic.Field.Shared.ExtendedInformation","type":"module"},{"doc":"","ref":"Packmatic.Field.Shared.ExtendedInformation.html#t:t/0","title":"Packmatic.Field.Shared.ExtendedInformation.t/0","type":"type"},{"doc":"Represents the Extended Timestamp Extra Field, which is emitted in both Local and Central File Headers. The field is emitted with only the modification time, in seconds since UNIX epoch (1 January, 1970). Structure Shared Extended Timestamp (UTc) Size Content 2 bytes Signature 2 bytes Size of Rest of Field (Bytes) 1 byte Flags 4 bytes Modification Time (Seconds since UNIX Epoch) Notes The flag value is 1 , representing that Modification Time is set.","ref":"Packmatic.Field.Shared.ExtendedTimestamp.html","title":"Packmatic.Field.Shared.ExtendedTimestamp","type":"module"},{"doc":"","ref":"Packmatic.Field.Shared.ExtendedTimestamp.html#t:t/0","title":"Packmatic.Field.Shared.ExtendedTimestamp.t/0","type":"type"},{"doc":"The Shared Timestamp field is emitted in both Local and Central File Headers, and is emitted in DOS (FAT) format. The time and date components are represented as little-endian, 16-bit integers, though they are first built separately Structure Size Content 2 bytes Hour (5 bits), Minute (6 bits), Second / 2 (5 bits) 2 bytes Year Since 1980 (7 bits), Month (4 bits), Day (5 bits) Notes See Erlang/OTP’s zip module: dos_date_time_to_datetime/2 . If the Timestamp given is prior to midnight, 1 January, 1980, it is also coerced to midnight, 1 January, 1980. If the Timestamp is on or after midnight, 1 January, 2108, then it can no longer be correctly represented within the limitations of the underlying field, and so is coerced to the previous representatable tick: 23:58, 31 December, 2107.","ref":"Packmatic.Field.Shared.Timestamp.html","title":"Packmatic.Field.Shared.Timestamp","type":"module"},{"doc":"","ref":"Packmatic.Field.Shared.Timestamp.html#t:t/0","title":"Packmatic.Field.Shared.Timestamp.t/0","type":"type"},{"doc":"Contains convenience functions which can be used to easily integrate a Zip stream with Plug-using applications such as Phoenix.","ref":"Packmatic.Conn.html","title":"Packmatic.Conn","type":"module"},{"doc":"Convenience function which sends the stream to the conn. The content of the Stream will be sent with an appropriately configured Content-Disposition response header (as attachment), and the name provided will be encoded for maximum compatibility with browsers. Examples stream |&gt; Packmatic.Conn . send_chunked ( conn , &quot;download.zip&quot; )","ref":"Packmatic.Conn.html#send_chunked/3","title":"Packmatic.Conn.send_chunked/3","type":"function"},{"doc":"","ref":"Packmatic.StreamError.html","title":"Packmatic.StreamError","type":"exception"},{"doc":"Packmatic Packmatic generates Zip streams by aggregating File or URL Sources. By using a Stream, the caller can compose it within the confines of Plug’s request/response model and serve the content of the resultant Zip archive in a streaming fashion. This allows fast delivery of a Zip archive consisting of many disparate parts hosted in different places, without having to first spool all of them to disk. The generated archive uses Zip64, and works with individual files that are larger than 4GB. See the Compatibility section for more information. Design Rationale Installation Usage Source Types Events Notes","ref":"readme.html","title":"Packmatic","type":"extras"},{"doc":"Problem In modern Web applications, content is often stored away from the host, such as Amazon S3, to enable better scalability and resiliency, as data would not be lost should a particular host go down. However, one disadvantage of this design is that data has become more spread-out, so previously simple operations such as creation of a Zip archive consisting of various files is now more complicated. In a hypothetical scenario where the developer opts for synchronous generation, the process may spend too long preparing the constituent files, which breaks certain connection proxies which expect a reasonable time to first byte. And in any case, disk utilisation will be high, since not only does each constituent file need to be on disk, there must be free space that will hold the archive temporarily as well. In another hypothetical scenario, where the developer opts for asynchronous generation, spooling still has to happen and further infrastructure for background job execution, user notification, temporary storage, etc. is now needed, which increases complexity. Solution The overall goal for Packmatic is to deliver a modern Zip streamer, which is able to assemble a Zip file from various sources, such as HTTP(S) downloads or locally generated temporary files, on-the-fly, so as to minimise waiting time, reduce resource consumption and elevate customer happiness. Packmatic solves the problem by separating the concerns of content definition (specifying what content should be retrieved, and from where), data retrieval (getting the content from the given source), and content presentation (creation of the Archive bundle). With Packmatic, the developer first specifies a list of Source Entries, which indicates where content can be obtained, and in what way: downloaded from the Internet, read from the filesystem, or dynamically resolved when the archive is being built, etc. This information is then wrapped in a Stream, which is consumed to send out chunked responses. When the Stream is started, it spins up an Encoder process, which works in a staged, iterative manner. This allows downloads to start almost immediately, which enhances customer happiness. The Encoder consumes the manifest as the Stream is consumed. In other words, the Stream starts producing data only when the client starts reading, and only as fast as the client reads, which reduces the amount of networking and local storage needed to start the process. Most of the acquisition work is interleaved, and back-pressured by client connection speed. Benefits Since Packmatic only streams what it needs, and does not spool content locally, it is possible to produce large archives whose constituent files do not fit on one host. This is an advantage over conventional processes, where even in a synchronous scenario, all constituent files and the archive must be spooled to disk. This design enhances developer happiness as well, since by using a synchronous, iterative mode of operation, there is no longer any need to spool constituent files in the archive on disk; there is also no longer any need to build technical workarounds in order to compensate for archive preparation delays. With Packmatic, both the problem and the solution is drastically simplified. The user clicks “download”, and the archive starts downloading. This is how it should be.","ref":"readme.html#design-rationale","title":"Packmatic - Design Rationale","type":"extras"},{"doc":"To install Packmatic, add the following line in your application’s dependencies: defp deps do [ { :packmatic , &quot;~&gt; 1.2.0&quot; } ] end","ref":"readme.html#installation","title":"Packmatic - Installation","type":"extras"},{"doc":"The general way to use Packmatic within your application is to generate a Stream dynamically by passing a list of Source Entries directly to Packmatic.build_stream/2 . This gives you a standard Stream which you can then send it off for download with help from Packmatic.Conn.send_chunked/3 . If you need more control, for example if you desire context separation, or if you wish to validate that the entries are valid prior to vending a Stream, you may generate a Packmatic.Manifest struct ahead of time, then pass it to Packmatic.build_stream/2 at a later time. See Packmatic.Manifest for more information. In either case, the Stream is powered by Packmatic.Encoder , which consumes the Entries within the Manifest iteratively as the Stream is consumed, at the pace set by the client’s download connection. Building the Stream with Entries The usual way to construct a Stream is as follows. entries = [ [ source : { :file , &quot;/tmp/hello.pdf&quot; } , path : &quot;hello.pdf&quot; ] , [ source : { :file , &quot;/tmp/world.pdf&quot; } , path : &quot;world.pdf&quot; , timestamp : DateTime . utc_now ( ) ] , [ source : { :url , &quot;https://example.com/foo.pdf&quot; } , path : &quot;foo/bar.pdf&quot; ] ] stream = Packmatic . build_stream ( entries ) As you can see, each Entry used to build the Stream (under source: ) is a keyword list, which concerns itself with the source, the path, and optionally a timestamp: source: represents a 2-arity tuple, representing the name of the Source and its Initialisation Argument. This data structure specifies the nature of the data, and how to obtain its content. path: represents the path in the Zip file that the content should be put under; it is your own responsibility to ensure that paths are not duplicated (see the Notes for an example). timestamp: is optional, and represents the creation/modification timestamp of the file. Packmatic emits both the basic form (DOS / FAT) of the timestamp, and the Extended Timestamp Extra Field which represents the same value with higher precision and range. Packmatic supports reading from any Source which conforms to the Packmatic.Source behaviour. To aid adoption and general implementation, there are built-in Sources as well; this is documented under Source Types . Building a Manifest If you wish, you can use the Packmatic.Manifest module to build a Manifest ahead-of-time, in order to validate the Entries prior to vending the Stream. Manifests can be created iteratively by calling Packmatic.Manifest.prepend/2 against an existing Manifest, or by calling Packmatic.Manifest.create/1 with a list of Entries created elsewhere. For more information, see Packmatic.Manifest . Specifying Error Behaviour By default, Packmatic fails the Stream when any Entry fails to process for any reason. If you desire, you may pass an additional option to Packmatic.build_stream/2 in order to modify this behaviour: stream = Packmatic . build_stream ( entries , on_error : :skip ) Writing Stream to File You can use the standard Stream.into/2 call to operate on the Stream: stream |&gt; Stream . into ( File . stream! ( file_path , [ :write ] ) ) |&gt; Stream . run ( ) Writing Stream to Conn (with Plug) You can use the bundled Packmatic.Conn module to send a Packmatic stream down the wire: stream |&gt; Packmatic.Conn . send_chunked ( conn , &quot;download.zip&quot; ) When writing the stream to a chunked Plug.Conn , Packmatic automatically escapes relevant characters in the name and sets the Content Disposition to attachment for maximum browser compatibility under intended use.","ref":"readme.html#usage","title":"Packmatic - Usage","type":"extras"},{"doc":"Packmatic has default Source types that you can use easily when building Manifests and/or Streams: File, representing content on disk, useful when the content is already available and only needs to be integrated. See Packmatic.Source.File . URL, representing content that is available remotely. Packmatic will run a chunked download routine to incrementally download and archive available chunks. The initialisation arguments for URL Sources can be referred with 2-arity or 3-arity tuples: {url, headers, options} or {url, options} . See Packmatic.Source.URL . Stream, representing content that is generated by a Stream (perhaps from other libraries) which can be incrementally consumed and incorporated in the archive. Anything that conforms to the Enumerable protocol, for example a List containing binaries, can be passed as well. Random, representing randomly generated bytes which is useful for testing. See Packmatic.Source.Random . Dynamic, representing a dynamically resolved Source, which is ultimately fulfilled by pulling content from either a File or an URL. If you have any need to inject a dynamically generated file, you may use this Source type to do it. This also has the benefit of avoiding expensive computation work in case the customer abandons the download midway. See Packmatic.Source.Dynamic . These Streams can be referred in 2-arity tuples by their internal aliases and the respective initialisation arguments: {:file, &quot;/tmp/hello/pdf&quot;} . {:url, &quot;https://example.com/hello/pdf&quot;} . {:stream, [&lt;&lt;0&gt;&gt;, &lt;&lt;1&gt;&gt;]} . {:random, 1048576} . {:dynamic, fn -&gt; {:ok, {:random, 1048576}} end} . Alternatively, they can also be referred by module names: {Packmatic.Source.File, &quot;/tmp/hello/pdf&quot;} . {Packmatic.Source.URL, &quot;https://example.com/hello/pdf&quot;} . {Packmatic.Source.Stream, enum} . {Packmatic.Source.Random, 1048576} . {Packmatic.Source.Dynamic, fn -&gt; {:ok, {:random, 1048576}} end} . Dynamic &amp; Custom Sources If you have an use case where you wish to dynamically generate the content that goes into the archive, you may either use the Dynamic source or implement a Custom Source. For example, if the amount of dynamic computation is small, but the results are time-sensitive, like when you already have Object IDs and just need to pre-sign URLs, you can use a Dynamic source with a curried function: { :dynamic , MyApp.Packmatic . build_dynamic_fun ( object_id ) } If you have a different use case, for example if you need to pull data from a FTP server (which uses a protocol that Packmatic does not have a bundled Source to work with), you can implement a module that conforms to the Packmatic.Source behaviour, and pass it: { MyApp.Packmatic.Source.FTP , &quot;ftp://example.com/my.docx&quot; } See Packmatic.Source for more information.","ref":"readme.html#source-types","title":"Packmatic - Source Types","type":"extras"},{"doc":"The Encoder can be configured to emit events in order to enable feedback elsewhere in your application, for example: entries = [ [ source : { :file , &quot;/tmp/hello.pdf&quot; } , path : &quot;hello.pdf&quot; ] , [ source : { :file , &quot;/tmp/world.pdf&quot; } , path : &quot;world.pdf&quot; , timestamp : DateTime . utc_now ( ) ] , [ source : { :url , &quot;https://example.com/foo.pdf&quot; } , path : &quot;foo/bar.pdf&quot; ] ] entries_count = length ( entries ) entries_completed_agent = Agent . start ( fn -&gt; 0 end ) handler_fun = fn event -&gt; case event do % Packmatic.Event.EntryCompleted { } -&gt; count = Agent . get_and_update ( entries_completed_agent , &amp; &amp;1 + 1 ) IO . puts &quot; \#{ count } of \#{ entries_count } encoded&quot; % Packmatic.Event.StreamEnded { } -&gt; :ok = Agent . stop ( entries_completed_agent ) :ok _ -&gt; :ok end end stream = Packmatic . build_stream ( entries , on_event : handler_fun ) See documentation for Packmatic.Event for a complete list of Event types. Compatibility Windows Windows Explorer (Windows 10): OK 7-Zip: OK macOS Finder (High Sierra): OK The Unarchiver (High Sierra): NG Erlang/OTP ( :zip ): NG Known Issues Ability to switch between STORE and DEFLATE modes is under investigation. Currently all archives are generated with DEFLATE, although for some use cases, this only increases the size of the archive, as content within each file is already compressed. Can not change compression level. Explicit directory generation is not supported; each entry must be a file. This is pending further investigation. Handling of 0-length files from sources is pending further investigation. Currently, they do get journaled but this may change in the future. Future Enhancements We will consider adding ability to resume downloads with range queries, so if the URL Source fails we can attempt to resume encoding from where we left off. In practice, if there are large files that are archived repeatedly, they will be pulled repeatedly and there is currently no explicit ability to cache them. We expect that in these cases an optimisation by the application developer may be to temporarily spool these hot constituents on disk and use a Dynamic Source to wrap around the cache layer (which would emit a File Source if the file is available locally, and act accordingly otherwise).","ref":"readme.html#events","title":"Packmatic - Events","type":"extras"},{"doc":"As with any user-generated content, you should exercise caution when building the Manifest, and ensure that only content that the User is entitled to retrieve is included. Due to design limitations, when downloading resources via a HTTP(S) connection, should the connection become closed halfway, a partial representation may be embedded in the output. A future release of this library may correct this in case the Content-Length header is present in the output, and act accordingly. You must ensure that paths are not duplicated within the same Manifest. You can do this by first building a list of sources and paths, then grouping and numbering them as needed with Enum.group_by/3 . Given entries are {source, path} tuples, you can do this: annotate_fun = fn entries -&gt; for { { source , path } , index } &lt;- Enum . with_index ( entries ) do path_components = Path . split ( path ) { path_components , [ filename ] } = Enum . split ( path_components , - 1 ) extname = Path . extname ( filename ) basename = Path . basename ( filename , extname ) path_components = path_components ++ [ &quot; \#{ basename } ( \#{ index + 1 } ) \#{ extname } &quot; ] { source , Path . join ( path_components ) } end end duplicates_fun = fn [ _ | [ _ | _ ] ] = entries -&gt; annotate_fun . ( entries ) entries -&gt; entries end entries |&gt; Enum . group_by ( &amp; elem ( &amp;1 , 1 ) ) |&gt; Enum . flat_map ( &amp; duplicates_fun . ( elem ( &amp;1 , 1 ) ) ) You must ensure that paths conform to the target environment of your choice, for example macOS and Windows each has its limitations regarding how long paths can be. The Stream emitted by Packmatic contains IO Data basically lists of binaries or themselves. To use the Stream with other libraries such as ex_aws_s3, you will have to write a transformer which converts the IO Lists to binaries. See the guest post. Guest Post: Using Packmatic with AWS S3 Multipart Uploads The following Guest Post is authored by Kyle Steger at Driver Technologies. Our company develops a dashcam application. We sync loads of sensor data (locations, acceleration, orientation, attitudes, ect) to our cloud platform while folks are driving with the application. This data is voluminous and infrequently queried so the natural place to store it is in S3. We allow our users to download exports of their data through our portal. We also share anonymized and aggregated information with some of our partners, who are typically interested in smaller segments of data around specific scenarios identified throughout a drive (think: hard braking, near collision, emergency vehicles). Today, that process could be summarized in a few steps: Fetch sensor data from S3 Filter/transform the data in memory Write the result to an export directory on disk Create an archive of the directory using :erlang.zip and write to disk Pray everything worked as it should Stream the archive to S3 Remove all artifacts from disk Packmatic allows us to remove steps 3, 4, 5, and 7 from our export flow. Since Packmatic works with streams, we create sources that are responsible for filtering the data and emitting the buffer that Packmatic passes off to the ExAws.S3.upload function. The only thing that I had to implement to support this was an intermediate Stream.transform on the Packmatic.build_stream() . Example below. The default chunk size is 10mb; S3 has a limit of 1000 upload chunks. Ideally this value is estimated based on the export. This function takes a stream generated by Packmatic.build_stream() and chunks the emitted values into sizes/shapes that ExAws.S3.upload can handle. @chunk_size 10_485_760 @spec chunk_packmatic_stream ( Enumerable . t ( ) ) :: Enumerable . t ( ) defp chunk_packmatic_stream ( stream ) do Stream . transform ( stream , _start_fun = fn -&gt; { _buffer = &quot;&quot; , _size = 0 } end , _reducer = fn elem , { buffer , size } when size &gt;= @chunk_size -&gt; new_buffer = IO . iodata_to_binary ( elem ) { [ buffer ] , { new_buffer , byte_size ( new_buffer ) } } elem , { buffer , size } -&gt; buffer_appendage = IO . iodata_to_binary ( elem ) { [ ] , { buffer &lt;&gt; buffer_appendage , size + byte_size ( buffer_appendage ) } } end , _last_fun = fn { buffer , _size } -&gt; { [ buffer ] , { &quot;&quot; , 0 } } end , _after_fun = fn _acc -&gt; :ok end ) end","ref":"readme.html#notes","title":"Packmatic - Notes","type":"extras"},{"doc":"During design and prototype development of this library, the Author has drawn inspiration from the following implementations, and therefore thanks all contributors for their generosity: ctrabant/fdzipstream dgvncsz0f/zipflow The Author wishes to thank the following individuals: Alvise Susmel for proposing and testing Encoder Events Christoph Geschwind for highlighting the need for explicit cleanup logic Derek Kraan &amp; Kyle Steger for Stream debugging","ref":"readme.html#acknowledgements","title":"Packmatic - Acknowledgements","type":"extras"},{"doc":"https://users.cs.jmu.edu/buchhofp/forensics/formats/pkzip.html https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT","ref":"readme.html#reference","title":"Packmatic - Reference","type":"extras"},{"doc":"Changelog All notable changes to this project will be documented in this file. The format is based on Keep a Changelog , and this project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Changed Changed ibrowse to hackney","ref":"changelog.html#1-2-0-7-january-2024","title":"Changelog - [1.2.0] — 7 January 2024","type":"extras"},{"doc":"Changed Revised handling of URL sources. Updated handling of SSL connections based on OTP 25 changes. Added SSL verify_peer by default. Added CA Certificates handling. Updated ibrowse to 4.4.2. Updated httpotion to 3.2.0.g","ref":"changelog.html#1-1-4-12-december-2023","title":"Changelog - [1.1.4] — 12 December 2023","type":"extras"},{"doc":"Changed Revised handling of Stream sources.","ref":"changelog.html#1-1-3-5-september-2023","title":"Changelog - [1.1.3] — 5 September 2023","type":"extras"},{"doc":"Changed Revised documentation and fixed formatting issues. Updated typespecs. Updated development &amp; test dependencies.","ref":"changelog.html#1-1-2-5-july-2021","title":"Changelog - [1.1.2] — 5 July 2021","type":"extras"},{"doc":"Added Added Packmatic.Source.Stream . Added support for Streams that output IO Lists. Updated Packmatic.Source . Added ability for any Source to use any data type as its Source State. Added ability for any Source to return an updated Source State with new data. Changed Revised Packmatic.Source.Dynamic . Removed custom resolver; any entry notation now accepted. Updated development &amp; test dependencies.","ref":"changelog.html#1-1-1-2-march-2021","title":"Changelog - 1.1.1 — 2 March 2021","type":"extras"},{"doc":"Added Added support for custom Sources. Any module which implements Packmatic.Source can be used as a Source. Added support for Encoder Events. Added the on_event option to the Encoder which can be used to receive events. See documentation for Packmatic.Event . Changed Revised Packmatic.Source . Added callback validate/1 for entry validation. Revised Packmatic.Manifest.Entry . Moved validation of Initialisation Arguments to Sources. Revised Packmatic.Source.File . Added explicit cleanup logic. Revised Packmatic.Source.URL . Added explicit cleanup logic. Fixed Revised Packmatic.Encoder . Fixed acceptance of IO Lists, in case of custom Sources returning these instead of binaries.","ref":"changelog.html#1-1-0-3-october-2020","title":"Changelog - 1.1.0 — 3 October 2020","type":"extras"},{"doc":"Changed Revised Manifests handling. Revised Manifests so they are validated once, during creation. Empty Manifests are now invalid. Manifests can be valid or invalid depending on their entries. See documentation for Packmatic.Manifest . Revised Encoder to not re-validate Manifests. Encoder halts immediately if given an invalid Manifest. Removed Packmatic.Validator.validate_each/1 . Since the Manifest is validated as it is built, this function is no longer useful. Revised tests. Revised test on “no entries” case for invalid Manifest. Added simple Manifest test with examples. Made the top-level PackmaticTest asynchronous. Revised Sources handling. Eliminated duplicative types with code generation. Known Sources are referred by name. The Source Entry type is generated based on the names. Revised individual Source modules. Standardised nomenclature ( init_arg , init_result , etc) for type handling. Revised Manifest types. Removed aliasing of Manifest.Entry.t() . Renamed Manifest.entry_keyword() to Manifest.Entry.proplist() . Added further documentation on how Dynamic Sources work. Added documentation within Packmatic.Source.Dynamic . Added ExDoc test for inlined snippets.","ref":"changelog.html#1-0-0-18-november-2019","title":"Changelog - 1.0.0 — 18 November 2019","type":"extras"},{"doc":"Added Initial Release.","ref":"changelog.html#0-1-0-30-october-2019","title":"Changelog - 0.1.0 — 30 October 2019","type":"extras"}]